# Copyright 1998 by Marco d'Itri <md@linux.it>
# You can use this software under the terms of the GPL. If we meet some day,
# and you think this stuff is worth it, you can buy me a beer in return.

package News::ICC;

use strict;
use vars qw($VERSION);

use Carp;
use Socket;
use POSIX qw(:errno_h :fcntl_h);

$VERSION="0.2";

#require 'inndcomm.ph';	# h2ph does not works?
sub ICC_PROTOCOL_1 () {'a'}
sub SC_SEP () {"\001"}
sub SC_CANCEL () {'c'}
sub SC_GO () {'h'}
sub SC_PAUSE () {'m'}
sub SC_RESERVE () {'z'}
sub SC_SHUTDOWN () {'q'}
sub SC_XABORT () {'x'}
sub SC_XEXEC () {'y'}

sub new {
	my $class=shift;
	croak 'you must require innshellvars.pl' if not defined $inn::serverpid;
	my $self={
		failure => '', timeout => 120, pending => 0, nonblock => 0,
	};
	bless $self, $class;
	return $self;
	# lint food:
	$inn::tempsockdir=$inn::newscontrol=$inn::tempsock;
}

sub open {		# Get ready to talk to the server
	my $self=shift;
	my $mysock="$inn::tempsockdir/$inn::tempsock.tmp";
	if (not open(FILE, ">$mysock")) {
		$self->{failure}='open';
		return 0;
	}
	if (not unlink($mysock)) {
		$self->{failure}='unlink';
		return 0;
	}
	if (not socket(ICCSOCKET, AF_UNIX, SOCK_DGRAM, 0)) {
		$self->{failure}='socket';
		return 0;
	}
	my $oldumask=umask(0);
	if (not bind(ICCSOCKET, pack_sockaddr_un($mysock))) {
		umask($oldumask);
		$self->{failure}='bind';
		return 0;
	}
	umask($oldumask);
	if ($self->{nonblock}) {
		if (not fcntl(ICCSOCKET, F_SETFL, O_NONBLOCK)) {
			$self->{failure}='fcntl';
			return 0;
		}
	}
	
	$self->{fd}=\*ICCSOCKET;
	$self->{sockname}=$mysock;
	$self->{serv}=pack_sockaddr_un($inn::newscontrol);
	return 1;
}

sub close {			# Close down
	my $self=shift;
	if (not close $self->{fd}) {
		$self->{failure}='close';
		return 0;
	}
	if (not unlink($self->{sockname})) {
		$self->{failure}='unlink';
		return 0;
	}
	return 1;
}

sub settimeout {	# Set the timeout
	my ($self, $t)=@_;
	$self->{timeout}=$t;
}

sub serverpid {		# Get the server's pid
	my $pid=1;
	if (open(FILE, $inn::serverpid)) {
		$pid=<FILE>;
		close FILE;
	}
	return $pid;
}

sub serveralive {	# See if the server is still there. Assume yes.
	my ($self, $pid)=@_;
	return 1 if (kill(0, $pid) > 0);
	return 0;
}

sub command {		# Send an arbitrary command to the server
	my ($self, $cmd, $args, $reply)=@_;

	croak 'bad arguments' if not defined $args;
	my $pid=$self->serverpid;
	if (not $self->serveralive($pid)) {
		$self->{failure}='dead server';
		return -1;
	}

	# Format the message
	my $msg=$self->{sockname} . SC_SEP . $cmd;
	foreach (@$args) {
		$msg.=SC_SEP . $_;
	}
	$msg.="\0"; # terminate the message

	# ana* is char, network short, chars
	$msg=pack('ana*', ICC_PROTOCOL_1, length($msg)+3, $msg);

	if (not send($self->{fd}, $msg, 0, $self->{serv})) {
		$self->{failure}='sendto';
		return -1;
	}

	return 0 if ($cmd eq SC_SHUTDOWN or $cmd eq SC_XABORT or $cmd eq SC_XEXEC);
	$self->{pending}++;

	my $rin=''; my $rsp;
	vec($rin, fileno($self->{fd}), 1)=1;
	LOOP: while($self->{pending}) {
#print "----->before select\n";
		# Wait for the reply (only if the socket is blocking)
		if ($self->{nonblock} or select($rin, undef, $rin, $self->{timeout})>0){
			while($self->{pending}) { # Server reply is there
#print "----->before recv\n";
				if (not recv($self->{fd}, $rsp, 4096, 0)) {
					last LOOP if ($! == EWOULDBLOCK);
					$self->{failure}='read';
					return -1;
				}
				$self->{pending}--;
#print "----->$rsp\n";
			}
		} else {
			if ($self->{timeout}) {	# No data -- if we timed out, return
				$self->{failure}='timeout';
				return -1;
			}
			if (not $self->serveralive($pid)) {
				$self->{failure}='dead server';
				return -1;
			}
		}
	}
	# $rsp is empty if there is not data to read from a non blocking socket
	return 0 if (not defined $rsp or $rsp eq '');

	my($protocol, $length, $rep)=unpack('ana*', $rsp);
	if ($protocol ne ICC_PROTOCOL_1) {
		$self->{failure}='protocol mismatch';
		return -1;
	}

	$$reply=$rep if (defined $reply);
	# Parse the rest of the reply; expected to be like <exitcode><space><text>
	my ($r, $t)=split(/ /, $rep, 2);
	return wantarray ? ($r, $t) : ($r || 0);
}

sub cancel {
	my ($self, $msgid)=@_;
	return $self->command(SC_CANCEL, [$msgid]);
}

sub go {
	my ($self, $why)=@_;
	return $self->command(SC_GO, [$why]);
}

sub pause {
	my ($self, $why)=@_;
	return $self->command(SC_PAUSE, [$why]);
}

sub reserve {
	my ($self, $why)=@_;
	return $self->command(SC_RESERVE, [$why]);
}

sub nonblocking {
	my ($self, $p)=@_;
	return $self->{nonblock}=($p || 1);
}

sub failure { $_[0]->{'failure'} } 

__END__

=head1 NAME

News::ICC - The Innd Control Command protocol

=head1 SYNOPSIS

	use News::ICC;
	my $i=new News::ICC;
	$i->open or die $i->failure.": $!";
	if ($i->cancel('<123456@abcdef>') < 0) {
		die $i->failure.": $!";
	}
	$i->close or die $i->failure.": $!";

=head1 DESCRIPTION

This is a perl implementation of the inndcomm.c library supplied by INN
in libinn. All functions of inndcomm are supported.

I implemented a new non blocking mode that can be enabled by calling the
nonblocking method B<BEFORE> calling the open method. When in non blocking
mode the command method discards all replies but the last.

The failuer method returns a string describing the last error.

=head1 AUTHOR

Marco d'Itri E<lt>md@linux.itE<gt>

=head1 LICENSE

You can use this software under the terms of the GPL. If we meet some day,
and you think this stuff is worth it, you can buy me a beer in return.

=head1 SEE ALSO

inndcomm(3), ctlinnd(8).

=cut

